const std = @import("std");

pub fn BlockingQueue(comptime max: usize, comptime T: type) type {
	return struct {
		buffer: [max]T = undefined,
		head: usize = 0,
		tail: usize = 0,
		mutex: std.Thread.Mutex = .{},
		not_empty: std.Thread.Condition = .{},
		not_full: std.Thread.Condition = .{},

		const Self = @This();

		pub fn enqueue(self: *Self, item: T) void {
			self.mutex.lock();

			defer self.mutex.unlock();

			while (((self.tail + 1) % max) == self.head) {
				self.not_full.wait(&self.mutex);
			}

			self.buffer[self.tail] = item;
			self.tail = (self.tail + 1) % max;

			self.not_empty.signal();
		}

		pub fn dequeue(self: *Self) T {
			self.mutex.lock();

			defer self.mutex.unlock();

			while (self.head == self.tail) {
				self.not_empty.wait(&self.mutex);
			}

			const item = self.buffer[self.head];

			self.buffer[self.head] = undefined;
			self.head = (self.head + 1) % max;

			self.not_full.signal();

			return item;
		}
	};
}

pub fn Future(comptime T: type) type {
	return struct {
		value: ?T = null,
		resolved: std.Thread.ResetEvent = .{},

		const Self = @This();

		pub fn get(self: *Self) *T {
			// TODO: Make async.
			self.resolved.wait();

			return &self.value.?;
		}

		pub fn resolve(self: *Self, value: T) bool {
			if (self.resolved.isSet()) {
				return false;
			}

			self.value = value;

			self.resolved.set();

			return true;
		}
	};
}
